
/*
 *
 * game.c - The game main loop
 *
 */

#include <stdio.h>

#include "common.h"
#include "random.h"
#include "game.h"
#include "video.h"
#include "audio.h"
#include "input.h"
#include "save.h"
#include "staticGeometry.h"
#include "collectables.h"
#include "robots.h"
#include "particles.h"
#include "baubles.h"
#include "hud.h"
#include "network.h"
#include "sprites.h"
#include "system.h"
#include "sfx.h"
#include "bgm.h"

// -----------------------------------------------------------------------
// variables
// -----------------------------------------------------------------------

// the game time
float gameTime = 0.0f;

// the run time
float runTime = 0.0f;

// the menu background scale
float menuBGScale = 38.0f;

// the key bits during previous iteration
int gameKeyBitsOld = 0;

// the game mode
int gameMode = GAME_MODE_SPLASH_1;//MENU_GAME_MP; // _SPLASH_1

// menu camera
float menuCameraX = 0.0f;
float menuCameraY = 0.0f;

// main menu selection
int menuSelectionMain = GAME_MENU_SELECTION_MAIN_QUICK_TEST;

// main menu selection center
float menuSelectionCenterMainX = 0.0f;
float menuSelectionCenterMainY = 0.0f;

// menu selection targets (MAIN)
float menuSelectionTargetsMain[] = {
	0.0f, // QUICK TEST
	7.6f, // MULTIPLAYER
	15.2f,  // CREDITS
	22.8f   // SETTINGS
};

// menu item coordinates (MAIN)
float menuItemCoordinatesMain[] = {
	0.0f,  0.0f, // QUICK TEST
	0.0f,  7.0f, // MULTIPLAYER
	0.0f, 14.0f, // CREDITS
	0.0f, 21.0f  // SETTINGS
};

// settings menu selection
int menuSelectionSettings = GAME_MENU_SELECTION_SETTINGS_BACK;

// settings menu selection center
float menuSelectionCenterSettingsX =  0.0f;
float menuSelectionCenterSettingsY = 15.4f;

// menu selection targets (SETTINGS)
float menuSelectionTargetsSettings[] = {
	-7.6f, // GFX LOW/HIGH
	 0.0f, // MUSIC ON/OFF
	 7.6f, // SFX ON/OFF
	15.4f  // BACK
};

// menu item coordinates (SETTINGS)
float menuItemCoordinatesSettings[] = {
	0.0f, -7.0f, // GFX LOW/HIGH
	0.0f,  0.0f, // MUSIC ON/OFF
	0.0f,  7.0f, // SFX ON/OFF
	0.0f, 14.0f  // BACK
};

// multiplayer menu selection
int menuSelectionMultiplayer = GAME_MENU_SELECTION_MULTIPLAYER_BACK;

// multiplayer menu selection center
float menuSelectionCenterMultiplayerX =  0.0f;
float menuSelectionCenterMultiplayerY = 30.0f;

// menu selection targets (MULTIPLAYER)
float menuSelectionTargetsMultiplayer[] = {
 -7.6f,  // NAME
	0.0f,  // SERVER
	22.0f, // CONNECT
	30.0f  // BACK
};

// menu item coordinates (MULTIPLAYER)
float menuItemCoordinatesMultiplayer[] = {
	0.0f, -7.0f, // NAME
	0.0f,  0.0f, // SERVER
	0.0f, 20.0f, // CONNECT
	0.0f, 27.0f  // BACK
};

// level selection mp menu selection
int menuSelectionLevelSelectionMP = GAME_MENU_SELECTION_LEVEL_SELECTION_MP_DISCONNECT;

// level selection mp menu selection center
float menuSelectionCenterLevelSelectionMPX =  0.0f;
float menuSelectionCenterLevelSelectionMPY = 36.5f;

// menu selection targets (LEVEL SELECTION MP)
float menuSelectionTargetsLevelSelectionMP[] = {
	10.0f, // LEVEL
	28.5f, // START
	36.5f  // DISCONNECT
};

// menu item coordinates (LEVEL SELECTION MP)
float menuItemCoordinatesLevelSelectionMP[] = {
	0.0f, 10.0f, // LEVEL
	0.0f, 26.0f, // START
	0.0f, 33.0f  // DISCONNECT
};

// the selection bar height (LEVEL SELECTION MP)
float menuSelectionBarDYLevelSelectionMPList[] = {
	26.0f, // LEVEL
	8.0f,  // START
	8.0f   // DISCONNECT
};

// the selection bar height (LEVEL SELECTION MP)
float menuSelectionBarDYLevelSelectionMP = 8.0f;

// virtual keyboard
char menuVirtualKeyboard[4][10] = {
	{ '1', '2', '3', '4', '5', '6', '7', '8', '9', '0' },
	{ 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J' },
	{ 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T' },
	{ 'U', 'V', 'W', 'X', 'Y', 'Z', '.', '.', '.', '@' }
};

// the virtual keyboard key coordinates
float menuVirtualKeyboardCoordinates[4][20];

// have the key coordinates been calculated?
int menuVirtualKeyboardInit = NO;

// virtual keyboard cursor coordinates
int menuVirtualKeyboardX = 9;
int menuVirtualKeyboardY = 3;

// virtual keyboard center
float menuVirtualKeyboardCenterY = VIRTUAL_KEYBOARD_REST_Y;

// virtual keyboard state
int menuVirtualKeyboardState = VIRTUAL_KEYBOARD_STATE_HIDE;

// player name
char playerName[16];
int  playerNameN = 0;

// server name
char serverName[16];
int  serverNameN = 0;

// the camera target
float cameraTargetX;
float cameraTargetY;

// the game mode we'll change into when the camera reaches its target
int cameraTargetGameMode;

// connect in connection screen?
int connectionScreenConnect;

// list of connected players
char playerNameList[4][16];
int  playerNameListN[4];

// the selected level
int menuLevelSelectionMPLevel = 0;

// the level sprites
int spritesLevel[8] = {
	VIDEO_SPRITE_MENU_LEVEL_1,
	VIDEO_SPRITE_MENU_LEVEL_2,
	VIDEO_SPRITE_MENU_LEVEL_3,
	VIDEO_SPRITE_MENU_LEVEL_4,
	VIDEO_SPRITE_MENU_LEVEL_5,
	VIDEO_SPRITE_MENU_LEVEL_6,
	VIDEO_SPRITE_MENU_LEVEL_7,
	VIDEO_SPRITE_MENU_LEVEL_8
};

// the game camera
float gameCameraX = 0.0f;
float gameCameraY = 0.0f;
float gameCameraZ = 0.0f;

// camera view mode
int gameCameraViewMode = 0;

// the camera angle
float gameCameraAngle = 0.0f;

// the player colors
int gamePlayerColors[] = {
	255,   0,   0, 0, // PLAYER 1 - RED
	124, 187, 110, 0, // PLAYER 2 - GREEN
	138, 110, 187, 0, // PLAYER 3 - VIOLET
	187, 151, 110, 0  // PLAYER 4 - BROWN
};

// -----------------------------------------------------------------------
// the tiles
// -----------------------------------------------------------------------

// the bases
int baseCoordinates[4*2] = {
	14, 14,
	22, 16,
	0, 0,
	0, 0,
};

// are the bases enabled?
int baseEnabled[4];

// the tile collision table
int tileCollisions[] = {
	NO,  // SAND
	NO,  // GRASS
	YES, // MOUNTAIN 1
	YES, // MOUNTAIN 2
	YES, // MOUNTAIN 3
	YES, // MOUNTAIN 4
	YES, // MOUNTAIN 5
	YES, // MOUNTAIN 6
	YES, // MOUNTAIN 7
	YES, // MOUNTAIN 8
	YES, // MOUNTAIN 9
	YES, // MOUNTAIN TOP
	NO,  // DARK SAND 1
	NO,  // DARK SAND 2
	YES, // MOUNTAIN INSIDE 1
	YES, // MOUNTAIN INSIDE 3
	YES, // MOUNTAIN INSIDE 7
	YES, // MOUNTAIN INSIDE 9
	NO,  // BASE
};

// the size of the tile map
int tileMapDX = 32;
int tileMapDY = 32;

// the tile size
float tileDX = 60.0f;
float tileDY = 60.0f;

// the tile map
int tileMap[32][32] = {
	{14, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,15 },
	{ 7, 2, 3, 3, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 5,11,11,11, 7, 0, 0, 0,12,13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 5,11,11,11, 7, 0, 0, 0,12,12,13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 5,11,11,11, 7, 0, 0, 0, 0,13,13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 8, 9, 9, 9,10, 0, 0, 0, 0,13,12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,13,13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,13,13, 0,13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,13,13,13, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5,11, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 9,10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{ 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 },
	{16, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,17 }
};

// -----------------------------------------------------------------------
// resources (COMMON)
// -----------------------------------------------------------------------

// the sprites (splash)
static int spritesSplash[] = {
	2, 2, 478, 194,   // GBAX2007
	2, 198, 478, 466  // TTURSAS
};

// the sprites (menu)
static int spritesMenu[] = {
   1, 0,  31, 42, // A
  33, 0,  66, 42, // B
  66, 0,  94, 42, // C
  96, 0, 124, 42, // D
 126, 0, 152, 42, // E
 155, 0, 181, 42, // F
 182, 0, 214, 42, // G
 217, 0, 249, 42, // H
 248, 0, 270, 42, // I
 272, 0, 297, 42, // J
 298, 0, 328, 42, // K
 329, 0, 354, 42, // L
 357, 0, 395, 42, // M
 396, 0, 427, 42, // N
 428, 0, 460, 42, // O
 463, 0, 493, 42, // P

   1, 46,  31, 90, // Q
  35, 46,  64, 90, // R
  67, 46,  99, 90, // S
 103, 46, 130, 90, // T
 133, 46, 161, 90, // U
 167, 46, 194, 90, // V
 197, 46, 234, 90, // W
 234, 46, 266, 90, // X
 266, 46, 297, 90, // Y
 297, 46, 331, 90, // Z

   0+8, 94,  30+8, 137, // 0
  36+8, 94,  57+8, 137, // 1
  58+8, 94,  88+8, 137, // 2
  91+8, 94, 120+8, 137, // 3
 123+8, 94, 152+8, 137, // 4
 154+8, 94, 182+8, 137, // 5
 185+8, 94, 215+8, 137, // 6
 218+8, 94, 246+8, 137, // 7
 258, 94, 292, 137, // 8
 295, 94, 328, 137, // 9

	250, 20, 257, 32, // COLON
	158, 21, 166, 36, // COMMA
	333, 56, 352, 89, // DOT
	361, 184, 384, 205, // MINUS
	262, 18, 284, 33, // ENTER
	64*0, 3*16, 64*1-1, 7*16, // LEVEL 1
	64*1, 3*16, 64*2-1, 7*16, // LEVEL 2
	64*2, 3*16, 64*3-1, 7*16, // LEVEL 3
	64*3, 3*16, 64*4-1, 7*16, // LEVEL 4
	64*4, 3*16, 64*5-1, 7*16, // LEVEL 5
	64*5, 3*16, 64*6-1, 7*16, // LEVEL 6
	64*6, 3*16, 64*7-1, 7*16, // LEVEL 7
	64*7, 3*16, 64*8-1, 7*16, // LEVEL 8

	 4, 149, 292, 277, // VIRTUAL KEYBOARD
	 304, 147, 426, 251, // LOGO
	 307, 262, 342, 296, // CURSOR
};

// the sprites (game)
static int spritesGame[] = {
	0, 0, 511, 511
};

// -----------------------------------------------------------------------
// resources (PSP)
// -----------------------------------------------------------------------

#ifdef PSP

// textures
extern unsigned char textureSplash_start[];
extern unsigned char textureMenu_start[];
extern unsigned char textureGame_start[];

// the texture atlases
static struct videoTextureAtlas videoTextureAtlases[] = {
	// SPLASH
	{
		"textureSplash.dtt",
		textureSplash_start,
		512,
		"textureSplash.pal",
		NULL,
		256,
		spritesSplash,
		2
	},
	// MENU
	{
		"textureMenu.dtt",
		textureMenu_start,
		512,
		"textureMenu.pal",
		NULL,
		256,
		spritesMenu,
		1
	},
	// GAME
	{
		"textureGame.dtt",
		textureGame_start,
		512,
		"textureGame.pal",
		NULL,
		256,
		spritesGame,
		1
	}
};

extern unsigned char sfxMenuClick_start[];

// the SFXs
static struct sfx sfxs[] = {
	{
		NULL,
		sfxMenuClick_start,
		13824/2
	}
};

// the BGMs
static char *bgms[] = {
	"bgmMenu.mp3"
};

#endif

// -----------------------------------------------------------------------
// resources (LINUX / WIN32)
// -----------------------------------------------------------------------

#ifndef PSP

// the texture atlases
static struct videoTextureAtlas videoTextureAtlases[] = {
	// SPLASH
	{
		"gfx/textureSplash.dtt",
		NULL,
		512,
		"gfx/textureSplash.pal",
		NULL,
		256,
		spritesSplash,
		2
	},
	// MENU
	{
		"gfx/textureMenu.dtt",
		NULL,
		512,
		"gfx/textureMenu.pal",
		NULL,
		256,
		spritesMenu,
		1
	},
	// GAME
	{
		"gfx/textureGame.dtt",
		NULL,
		512,
		"gfx/textureGame.pal",
		NULL,
		256,
		spritesGame,
		1
	}
};

// the SFXs
static struct sfx sfxs[] = {
	{
		"sfx/sfxMenuClick.wav",
		NULL,
		0
	}
};

// the BGMs
static char *bgms[] = {
	"bgm/bgmMenu.ogg"
};

#endif

// -----------------------------------------------------------------------
// init the game
// -----------------------------------------------------------------------

void gameInit(void) {

	int i;

	// overclock the systems that support overclocking (i.e., PSP)
	systemOverclock(YES);

	// set the BGM, SFX and texture lists
	audioSetBGMList(bgms, 1);
	audioSetSFXList(sfxs, 1);
	videoSetTextureList(videoTextureAtlases, 3);

	// start BGM
	audioPlayBGM(BGM_MENU);

	gameTime = 0.0f;

	// reset the names
	for (i = 0; i < 16; i++) {
		playerName[i] = 0;
		serverName[i] = 0;
	}
}

void gameGameInit(void) {

	int i;

	collectablesInit();
	robotsInit();
	baublesInit();
	hudInit();
	particlesInit();

	collectablesAdd(COLLECTABLE_TYPE_OIL, 1100.0f, 1100.0f, 100.0f);
	collectablesAdd(COLLECTABLE_TYPE_OIL, 1200.0f, 1100.0f, 140.0f);
	collectablesAdd(COLLECTABLE_TYPE_OIL, 1120.0f, 1140.0f, 0.0f);
	collectablesAdd(COLLECTABLE_TYPE_OIL, 1150.0f, 1110.0f, 20.0f);

	robotsAdd(ROBOT_TYPE_TOASTER, 1300.0f, 1120.0f, 240.0f, networkMyID, -1, ROBOT_STATE_IDLE);
	robotsAdd(ROBOT_TYPE_TOASTER, 1100.0f, 1000.0f, 20.0f, networkMyID, -1, ROBOT_STATE_IDLE);
	robotsAdd(ROBOT_TYPE_MICROWAVE, 1300.0f, 1020.0f, 240.0f, networkMyID, -1, ROBOT_STATE_IDLE);
	robotsAdd(ROBOT_TYPE_MICROWAVE, 1400.0f, 1420.0f, 240.0f, networkMyID, -1, ROBOT_STATE_IDLE);
	robotsAdd(ROBOT_TYPE_MICROWAVE, 1200.0f, 1200.0f, 240.0f, networkMyID, -1, ROBOT_STATE_IDLE);

	baublesAdd(BAUBLE_TYPE_ROCK, 1000.0f, 1000.0f, 0.0f, 1.0f, 20.0f);
	baublesAdd(BAUBLE_TYPE_ROCK, 1100.0f,  800.0f, -2.0f, 1.2f, 120.0f);
	baublesAdd(BAUBLE_TYPE_ROCK, 1120.0f,  790.0f, -4.0f, 2.0f, 220.0f);

	// set the bases
	for (i = 0; i < 4; i++)
		baseEnabled[i] = NO;

	for (i = 0; i < 2; i++) {
		tileMap[baseCoordinates[i*2 + 1]][baseCoordinates[i*2 + 0]] = TILE_BASE;
		baseEnabled[i] = YES;
	}

	gameCameraX = tileMapDX * tileDX * 0.5f;
	gameCameraY = tileMapDY * tileDY * 0.5f;
}

// -----------------------------------------------------------------------
// game iteration
// -----------------------------------------------------------------------

void gameGameIteration(float deltaTime) {

	float speed, dX, dY, tX, tY;
	int i;

	speed = deltaTime + (deltaTime*gameCameraZ)/1800.0f;

	// move the camera X, Y and Z
	if ((inputKeyBits & INPUT_KEY_BIT_LEFT) != 0)
		gameCameraX -= speed;
	if ((inputKeyBits & INPUT_KEY_BIT_RIGHT) != 0)
		gameCameraX += speed;
	if ((inputKeyBits & INPUT_KEY_BIT_UP) != 0)
		gameCameraY -= speed;
	if ((inputKeyBits & INPUT_KEY_BIT_DOWN) != 0)
		gameCameraY += speed;
	if ((inputKeyBits & INPUT_KEY_BIT_L) != 0)
		gameCameraZ += deltaTime * 0.8f;
	if ((inputKeyBits & INPUT_KEY_BIT_R) != 0)
		gameCameraZ -= deltaTime * 0.8f;

	// limit Z
	if (gameCameraZ < -1500.0f)
		gameCameraZ = -1500.0f;
	if (gameCameraZ > 0.0f)
		gameCameraZ = 0.0f;

	// run top-down view
	if (((inputKeyBits & INPUT_KEY_BIT_Y) != 0) && (gameKeyBitsOld & INPUT_KEY_BIT_Y) == 0) {
		gameCameraViewMode++;
		if (gameCameraViewMode == 3)
			gameCameraViewMode = 0;
	}

	if (gameCameraViewMode == 0) {
		gameCameraAngle -= deltaTime * 0.1f;
		if (gameCameraAngle < 0.0f)
			gameCameraAngle = 0.0f;
	}
	else if (gameCameraViewMode == 1) {
		gameCameraAngle += deltaTime * 0.1f;
		if (gameCameraAngle > 40.0f)
			gameCameraAngle = 40.0f;
	}
	else {
		gameCameraAngle += deltaTime * 0.1f;
		if (gameCameraAngle > 70.0f)
			gameCameraAngle = 70.0f;
	}

	// give a new target to the selected robot?
	if (robotSelected >= 0) {
		if ((inputKeyBits & INPUT_KEY_BIT_A) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_A) == 0) {
			// did we click near the base
			tX = gameCameraX;
			tY = gameCameraY + gameCameraAngle*(-1500.0f - gameCameraZ)*0.017f - 50.0f;

			fprintf(stderr, "tX = %.3f tY = %.3f\n", tX, tY);

			robots[robotSelected].state = ROBOT_STATE_GOTO_TARGET;

			for (i = 0; i < 4; i++) {
				if (baseEnabled[i] == NO)
					continue;

				dX = tX - (baseCoordinates[i*2 + 0] * tileDX);
				dY = tY - (baseCoordinates[i*2 + 1] * tileDY);

				if (dX*dX + dY*dY < 30.0f*30.0f) {
					// yep -> return to base
					tX = baseCoordinates[i*2 + 0] * tileDX;
					tY = baseCoordinates[i*2 + 1] * tileDY;
					robots[robotSelected].state = ROBOT_STATE_GOTO_BASE;
					robots[robotSelected].targetBase = i;
					break;
				}
			}

			robots[robotSelected].targetX = tX;
			robots[robotSelected].targetY = tY;
			robots[robotSelected].guardX = tX;
			robots[robotSelected].guardY = tY;
			//robots[robotSelected].health = 0;
		}
	}

	// select the next robot?
	if ((inputKeyBits & INPUT_KEY_BIT_B) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_B) == 0)
		robotsSelectNext();

	robotsIteration(deltaTime);
	baublesIteration(deltaTime);
	particlesIteration(deltaTime);
}

// -----------------------------------------------------------------------
// draw the game
// -----------------------------------------------------------------------

void gameDrawGame(void) {

	int i, j, tile;

	/*
	videoSetColor(248, 173, 27, 255);

	videoLoadIdentity();
	if (gameCameraAngle != 0.0f)
		videoRotate3f(gameCameraAngle, 0.0f, 0.0f);
	videoTranslate3f(0 - gameCameraX, 0 - gameCameraY, -1560.0f - gameCameraZ);
	videoDrawRectangle2D(8000.0f, 8000.0f);
	*/

	// set the camera rotations
	videoSetColor(255, 255, 255, 255);

	videoLoadIdentity();
	if (gameCameraAngle != 0.0f)
		videoRotate3f(gameCameraAngle, 0.0f, 0.0f);
	videoTranslate3f(-gameCameraX, -gameCameraY, -gameCameraZ);

	// draw the tiles
	for (i = 0; i < tileMapDY; i++) {
		for (j = 0; j < tileMapDX; j++) {
			videoMatrixPush();

			tile = tileMap[j][i];

			videoTranslate3f(i*tileDX, j*tileDY, -1550.0f);

			if (tile == TILE_SAND)
				videoDrawMeshTextured(verticesGameTileSand, verticesGameTileSandTriangles);
			else if (tile == TILE_DARK_SAND_1)
				videoDrawMeshTextured(verticesGameTileDarkSand1, verticesGameTileDarkSand1Triangles);
			else if (tile == TILE_DARK_SAND_2)
				videoDrawMeshTextured(verticesGameTileDarkSand2, verticesGameTileDarkSand2Triangles);
			else if (tile == TILE_MOUNTAIN_TOP)
				videoDrawMeshTextured(verticesGameTileMountainTop, verticesGameTileMountainTopTriangles);
			else if (tile == TILE_MOUNTAIN_1)
				videoDrawMeshTextured(verticesGameTileMountain1, verticesGameTileMountain1Triangles);
			else if (tile == TILE_MOUNTAIN_2) {
				videoDrawMeshTextured(verticesGameTileMountain4, verticesGameTileMountain4Triangles);
			}
			else if (tile == TILE_MOUNTAIN_3) {
				videoRotate3f(0.0f, 0.0f, 90.0f);
				videoDrawMeshTextured(verticesGameTileMountain1, verticesGameTileMountain1Triangles);
			}
			else if (tile == TILE_MOUNTAIN_4) {
				videoRotate3f(0.0f, 0.0f, 270.0f);
				videoDrawMeshTextured(verticesGameTileMountain4, verticesGameTileMountain4Triangles);
			}
			else if (tile == TILE_MOUNTAIN_5) {
				videoDrawMeshTextured(verticesGameTileMountain5, verticesGameTileMountain5Triangles);
			}
			else if (tile == TILE_MOUNTAIN_6) {
				videoRotate3f(0.0f, 0.0f, 90.0f);
				videoDrawMeshTextured(verticesGameTileMountain4, verticesGameTileMountain4Triangles);
			}
			else if (tile == TILE_MOUNTAIN_7) {
				videoRotate3f(0.0f, 0.0f, 270.0f);
				videoDrawMeshTextured(verticesGameTileMountain1, verticesGameTileMountain1Triangles);
			}
			else if (tile == TILE_MOUNTAIN_8) {
				videoRotate3f(0.0f, 0.0f, 180.0f);
				videoDrawMeshTextured(verticesGameTileMountain4, verticesGameTileMountain4Triangles);
			}
			else if (tile == TILE_MOUNTAIN_9) {
				videoRotate3f(0.0f, 0.0f, 180.0f);
				videoDrawMeshTextured(verticesGameTileMountain1, verticesGameTileMountain1Triangles);
			}
			else if (tile == TILE_MOUNTAIN_INSIDE_9) {
				videoDrawMeshTextured(verticesGameTileMountainInside9, verticesGameTileMountainInside9Triangles);
			}
			else if (tile == TILE_MOUNTAIN_INSIDE_7) {
				videoRotate3f(0.0f, 0.0f, 90.0f);
				videoDrawMeshTextured(verticesGameTileMountainInside9, verticesGameTileMountainInside9Triangles);
			}
			else if (tile == TILE_MOUNTAIN_INSIDE_1) {
				videoRotate3f(0.0f, 0.0f, 180.0f);
				videoDrawMeshTextured(verticesGameTileMountainInside9, verticesGameTileMountainInside9Triangles);
			}
			else if (tile == TILE_MOUNTAIN_INSIDE_3) {
				videoRotate3f(0.0f, 0.0f, 270.0f);
				videoDrawMeshTextured(verticesGameTileMountainInside9, verticesGameTileMountainInside9Triangles);
			}
			else if (tile == TILE_BASE) {
				videoDrawMeshTextured(verticesGameTileBase, verticesGameTileBaseTriangles);
			}

			videoMatrixPop();
		}
	}

	collectablesDraw();
	robotsDraw();
	baublesDraw();
	particlesDraw();
	hudDraw();
}

// -----------------------------------------------------------------------
// game iteration
// -----------------------------------------------------------------------

void gameIteration(float deltaTime) {

	int keyBits;
	float fY;

	if (deltaTime < 0.01f)
		deltaTime = 0.01f;

	// advance the run time
	runTime += deltaTime;

	// advance main timer
	gameTime += deltaTime;

	// read the key bits
	keyBits = inputKeyBits;

	switch (gameMode) {

		// -----------------------------------------------------------------------
		// SPLASH
		// -----------------------------------------------------------------------

	case GAME_MODE_SPLASH_1:
		// wait a little
		if (gameTime < 3000.0f)
			break;

		// move right, to GBAX2007
		menuCameraX += deltaTime * 0.1f;

		if (menuCameraX >= GAME_MENU_CENTER_SPLASH_1_X) {
			menuCameraX = GAME_MENU_CENTER_SPLASH_1_X;
			gameTime = 0.0f;
			gameMode = GAME_MODE_SPLASH_2;
		}

		break;

	case GAME_MODE_SPLASH_2:
		// wait a little with GBAX2007
		if (gameTime >= 2000.0f)
			gameMode = GAME_MODE_SPLASH_3;

		break;

	case GAME_MODE_SPLASH_3:
		// move right, to ttursas
		menuCameraX += deltaTime * 0.1f;

		if (menuCameraX >= GAME_MENU_CENTER_SPLASH_2_X) {
			menuCameraX = GAME_MENU_CENTER_SPLASH_2_X;
			gameTime = 0.0f;
			gameMode = GAME_MODE_SPLASH_4;
		}

		break;

	case GAME_MODE_SPLASH_4:
		// wait a little with ttursas
		if (gameTime >= 4000.0f)
			gameMode = GAME_MODE_SPLASH_5;

		break;

	case GAME_MODE_SPLASH_5:
		// move right, to the main menu
		menuCameraX += deltaTime * 0.1f;

		if (menuCameraX >= GAME_MENU_CENTER_SPLASH_3_X) {
			menuCameraX = GAME_MENU_CENTER_SPLASH_3_X;
			gameTime = 0.0f;
			gameMode = GAME_MODE_SPLASH_EXIT;
		}

		break;

	case GAME_MODE_SPLASH_EXIT:
		// move right, to the main menu
		menuCameraX += deltaTime * 0.4f;

		if (menuCameraX >= GAME_MENU_CENTER_MENU_MAIN_X) {
			menuCameraX = GAME_MENU_CENTER_MENU_MAIN_X;
			gameTime = 0.0f;
			gameMode = GAME_MODE_MENU_MAIN;
		}

		break;

		// -----------------------------------------------------------------------
		// MENU MAIN
		// -----------------------------------------------------------------------

	case GAME_MODE_MENU_MAIN:
		// manage input
		if ((keyBits & INPUT_KEY_BIT_UP) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_UP) == 0) {
			menuSelectionMain--;
			if (menuSelectionMain < 0)
				menuSelectionMain = 3;
		}
		if ((keyBits & INPUT_KEY_BIT_DOWN) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_DOWN) == 0) {
			menuSelectionMain++;
			if (menuSelectionMain > 3)
				menuSelectionMain = 0;
		}

		// move the menu selection
		fY = menuSelectionTargetsMain[menuSelectionMain];

		if (menuSelectionCenterMainY < fY) {
			menuSelectionCenterMainY += deltaTime * 0.1f;
			if (menuSelectionCenterMainY > fY)
				menuSelectionCenterMainY = fY;
		}
		else if (menuSelectionCenterMainY > fY) {
			menuSelectionCenterMainY -= deltaTime * 0.1f;
			if (menuSelectionCenterMainY < fY)
				menuSelectionCenterMainY = fY;
		}

		if (menuSelectionCenterMainY == fY) {
			// exit main menu?
			if (((keyBits & INPUT_KEY_BIT_START) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_START) == 0) ||
					((keyBits & INPUT_KEY_BIT_A) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_A) == 0)) {
				if (menuSelectionMain == GAME_MENU_SELECTION_MAIN_QUICK_TEST) {
					gameMode = GAME_MODE_MENU_GAME_MP;

					// reset the game
					gameGameInit();
				}
				else if (menuSelectionMain == GAME_MENU_SELECTION_MAIN_MULTIPLAYER)
					gameMode = GAME_MODE_MENU_MULTIPLAYER_IN;
				else if (menuSelectionMain == GAME_MENU_SELECTION_MAIN_CREDITS)
					gameMode = GAME_MODE_MENU_CREDITS_IN;
				else if (menuSelectionMain == GAME_MENU_SELECTION_MAIN_SETTINGS)
					gameMode = GAME_MODE_MENU_SETTINGS_IN;

				// play the click SFX
				audioPlaySFX(SFX_MENU_CLICK);
			}
		}

		break;

		// -----------------------------------------------------------------------
		// CREDITS
		// -----------------------------------------------------------------------

	case GAME_MODE_MENU_CREDITS_IN:
		// move down, to the credits screen
		menuCameraY += deltaTime * 0.4f;

		if (menuCameraY >= GAME_MENU_CENTER_MENU_CREDITS_Y) {
			menuCameraY = GAME_MENU_CENTER_MENU_CREDITS_Y;
			gameTime = 0.0f;
			gameMode = GAME_MODE_MENU_CREDITS;
		}

		break;

	case GAME_MODE_MENU_CREDITS_OUT:
		// move up, to the main menu screen
		menuCameraY -= deltaTime * 0.4f;

		if (menuCameraY <= GAME_MENU_CENTER_MENU_MAIN_Y) {
			menuCameraY = GAME_MENU_CENTER_MENU_MAIN_Y;
			gameTime = 0.0f;
			gameMode = GAME_MODE_MENU_MAIN;
		}

		break;

	case GAME_MODE_MENU_CREDITS:
		if (((keyBits & INPUT_KEY_BIT_START) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_START) == 0) ||
				((keyBits & INPUT_KEY_BIT_A) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_A) == 0)) {
			gameMode = GAME_MODE_MENU_CREDITS_OUT;

			// play the click SFX
			audioPlaySFX(SFX_MENU_CLICK);
		}

		break;

		// -----------------------------------------------------------------------
		// SETTINGS
		// -----------------------------------------------------------------------

	case GAME_MODE_MENU_SETTINGS_IN:
		// move up, to the settings screen
		menuCameraY -= deltaTime * 0.4f;

		if (menuCameraY <= GAME_MENU_CENTER_MENU_SETTINGS_Y) {
			menuCameraY = GAME_MENU_CENTER_MENU_SETTINGS_Y;
			gameTime = 0.0f;
			gameMode = GAME_MODE_MENU_SETTINGS;
		}

		break;

	case GAME_MODE_MENU_SETTINGS_OUT:
		// move down, to the main menu screen
		menuCameraY += deltaTime * 0.4f;

		if (menuCameraY >= GAME_MENU_CENTER_MENU_MAIN_Y) {
			menuCameraY = GAME_MENU_CENTER_MENU_MAIN_Y;
			gameTime = 0.0f;
			gameMode = GAME_MODE_MENU_MAIN;
		}

		break;

	case GAME_MODE_MENU_SETTINGS:
		// manage input
		if ((keyBits & INPUT_KEY_BIT_UP) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_UP) == 0) {
			menuSelectionSettings--;
			if (menuSelectionSettings < 0)
				menuSelectionSettings = 3;
		}
		if ((keyBits & INPUT_KEY_BIT_DOWN) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_DOWN) == 0) {
			menuSelectionSettings++;
			if (menuSelectionSettings > 3)
				menuSelectionSettings = 0;
		}

		// move the menu selection
		fY = menuSelectionTargetsSettings[menuSelectionSettings];

		if (menuSelectionCenterSettingsY < fY) {
			menuSelectionCenterSettingsY += deltaTime * 0.1f;
			if (menuSelectionCenterSettingsY > fY)
				menuSelectionCenterSettingsY = fY;
		}
		else if (menuSelectionCenterSettingsY > fY) {
			menuSelectionCenterSettingsY -= deltaTime * 0.1f;
			if (menuSelectionCenterSettingsY < fY)
				menuSelectionCenterSettingsY = fY;
		}

		if (menuSelectionCenterSettingsY == fY) {
			// exit main menu?
			if (((keyBits & INPUT_KEY_BIT_START) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_START) == 0) ||
					((keyBits & INPUT_KEY_BIT_A) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_A) == 0)) {
				if (menuSelectionSettings == GAME_MENU_SELECTION_SETTINGS_MUSIC) {
					if (audioBGM == YES)
						audioSetBGM(NO);
					else
						audioSetBGM(YES);
				}
				else if (menuSelectionSettings == GAME_MENU_SELECTION_SETTINGS_SFX) {
					if (audioSFX == YES)
						audioSetSFX(NO);
					else
						audioSetSFX(YES);
				}
				else if (menuSelectionSettings == GAME_MENU_SELECTION_SETTINGS_GFX) {
					if (videoQuality == VIDEO_QUALITY_LOW)
						videoSetQuality(VIDEO_QUALITY_HIGH);
					else
						videoSetQuality(VIDEO_QUALITY_LOW);
				}
				else if (menuSelectionSettings == GAME_MENU_SELECTION_SETTINGS_BACK)
					gameMode = GAME_MODE_MENU_SETTINGS_OUT;

				// play the click SFX
				audioPlaySFX(SFX_MENU_CLICK);
			}
		}

		break;

		// -----------------------------------------------------------------------
		// MULTIPLAYER
		// -----------------------------------------------------------------------

	case GAME_MODE_MENU_MULTIPLAYER_IN:
		// move right, to the multiplayer screen
		menuCameraX += deltaTime * 0.4f;

		if (menuCameraX >= GAME_MENU_CENTER_MENU_MULTIPLAYER_X) {
			menuCameraX = GAME_MENU_CENTER_MENU_MULTIPLAYER_X;
			gameTime = 0.0f;
			gameMode = GAME_MODE_MENU_MULTIPLAYER;
		}

		break;

	case GAME_MODE_MENU_MULTIPLAYER_OUT:
		// move left, to the main menu screen
		menuCameraX -= deltaTime * 0.4f;

		if (menuCameraX <= GAME_MENU_CENTER_MENU_MAIN_X) {
			menuCameraX = GAME_MENU_CENTER_MENU_MAIN_X;
			gameTime = 0.0f;
			gameMode = GAME_MODE_MENU_MAIN;
		}

		break;

	case GAME_MODE_MENU_MULTIPLAYER:
		// manage input
		if (menuVirtualKeyboardCenterY >= VIRTUAL_KEYBOARD_REST_Y) {
			if ((keyBits & INPUT_KEY_BIT_UP) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_UP) == 0) {
				menuSelectionMultiplayer--;
				if (menuSelectionMultiplayer < 0)
					menuSelectionMultiplayer = 3;
			}
			if ((keyBits & INPUT_KEY_BIT_DOWN) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_DOWN) == 0) {
				menuSelectionMultiplayer++;
				if (menuSelectionMultiplayer > 3)
					menuSelectionMultiplayer = 0;
			}
		}

		// move the menu selection
		fY = menuSelectionTargetsMultiplayer[menuSelectionMultiplayer];

		if (menuSelectionCenterMultiplayerY < fY) {
			menuSelectionCenterMultiplayerY += deltaTime * 0.1f;
			if (menuSelectionCenterMultiplayerY > fY)
				menuSelectionCenterMultiplayerY = fY;
		}
		else if (menuSelectionCenterMultiplayerY > fY) {
			menuSelectionCenterMultiplayerY -= deltaTime * 0.1f;
			if (menuSelectionCenterMultiplayerY < fY)
				menuSelectionCenterMultiplayerY = fY;
		}

		// move the virtual keyboard
		if (menuVirtualKeyboardState == VIRTUAL_KEYBOARD_STATE_SHOW) {
			menuVirtualKeyboardCenterY -= deltaTime * 0.1f;
			if (menuVirtualKeyboardCenterY < 16.0f)
				menuVirtualKeyboardCenterY = 16.0f;
		}
		else {
			menuVirtualKeyboardCenterY += deltaTime * 0.1f;
			if (menuVirtualKeyboardCenterY >= VIRTUAL_KEYBOARD_REST_Y)
				menuVirtualKeyboardCenterY = VIRTUAL_KEYBOARD_REST_Y;
		}

		if (menuVirtualKeyboardCenterY < VIRTUAL_KEYBOARD_REST_Y) {
			// the virtual keyboard is active
			if ((keyBits & INPUT_KEY_BIT_START) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_START) == 0) {
				menuVirtualKeyboardState = VIRTUAL_KEYBOARD_STATE_HIDE;

				// play the click SFX
				audioPlaySFX(SFX_MENU_CLICK);

				// save the new configuration
				saveSave();
			}
			else if ((keyBits & INPUT_KEY_BIT_A) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_A) == 0) {
				// play the click SFX
				audioPlaySFX(SFX_MENU_CLICK);

				if (menuVirtualKeyboardX == 9 && menuVirtualKeyboardY == 3) {
					menuVirtualKeyboardState = VIRTUAL_KEYBOARD_STATE_HIDE;

					// save the new configuration
					saveSave();
				}
				else if (menuSelectionMultiplayer == GAME_MENU_SELECTION_MULTIPLAYER_NAME) {
					if (playerNameN < 15)
						playerName[playerNameN++] = menuVirtualKeyboard[menuVirtualKeyboardY][menuVirtualKeyboardX];
				}
				else if (menuSelectionMultiplayer == GAME_MENU_SELECTION_MULTIPLAYER_SERVER) {
					if (serverNameN < 15)
						serverName[serverNameN++] = menuVirtualKeyboard[menuVirtualKeyboardY][menuVirtualKeyboardX];
				}
			}
			else if ((keyBits & INPUT_KEY_BIT_B) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_B) == 0) {
				// play the click SFX
				audioPlaySFX(SFX_MENU_CLICK);

				if (menuSelectionMultiplayer == GAME_MENU_SELECTION_MULTIPLAYER_NAME) {
					if (playerNameN > 0)
						playerName[--playerNameN] = 0;
				}
				else if (menuSelectionMultiplayer == GAME_MENU_SELECTION_MULTIPLAYER_SERVER) {
					if (serverNameN > 0)
						serverName[--serverNameN] = 0;
				}
			}
			else if ((keyBits & INPUT_KEY_BIT_UP) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_UP) == 0) {
				menuVirtualKeyboardY--;
				if (menuVirtualKeyboardY < 0)
					menuVirtualKeyboardY = 3;
			}
			else if ((keyBits & INPUT_KEY_BIT_DOWN) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_DOWN) == 0) {
				menuVirtualKeyboardY++;
				if (menuVirtualKeyboardY >= 4)
					menuVirtualKeyboardY = 0;
			}
			else if ((keyBits & INPUT_KEY_BIT_LEFT) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_LEFT) == 0) {
				menuVirtualKeyboardX--;
				if (menuVirtualKeyboardX < 0)
					menuVirtualKeyboardX = 9;
			}
			else if ((keyBits & INPUT_KEY_BIT_RIGHT) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_RIGHT) == 0) {
				menuVirtualKeyboardX++;
				if (menuVirtualKeyboardX >= 10)
					menuVirtualKeyboardX = 0;
			}
		}
		else {
			// the multiplayer menu eats the input
			if (menuSelectionCenterMultiplayerY == fY) {
				if (((keyBits & INPUT_KEY_BIT_START) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_START) == 0) ||
						((keyBits & INPUT_KEY_BIT_A) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_A) == 0)) {
					// play the click SFX
					audioPlaySFX(SFX_MENU_CLICK);

					if (menuSelectionMultiplayer == GAME_MENU_SELECTION_MULTIPLAYER_BACK)
						gameMode = GAME_MODE_MENU_MULTIPLAYER_OUT;
					else if (menuSelectionMultiplayer == GAME_MENU_SELECTION_MULTIPLAYER_CONNECT) {
						gameMode = GAME_MODE_MENU_CONNECTION_IN;
						cameraTargetX = GAME_MENU_CENTER_MENU_CONNECTION_X;
						cameraTargetGameMode = GAME_MODE_MENU_CONNECTION;
						connectionScreenConnect = YES;

						// initiate the connection to the server
						networkConnect();
						networkServerConnect(playerName, serverName);
					}
					else if (menuSelectionMultiplayer == GAME_MENU_SELECTION_MULTIPLAYER_NAME)
						menuVirtualKeyboardState = VIRTUAL_KEYBOARD_STATE_SHOW;
					else if (menuSelectionMultiplayer == GAME_MENU_SELECTION_MULTIPLAYER_SERVER)
						menuVirtualKeyboardState = VIRTUAL_KEYBOARD_STATE_SHOW;
				}
			}
		}
		break;

		// -----------------------------------------------------------------------
		// CONNECTION
		// -----------------------------------------------------------------------

	case GAME_MODE_MENU_CONNECTION:
		if (gameTime < 1500.0f)
			break;

		if (connectionScreenConnect == YES && networkServerConnectionStatus == NETWORK_SERVER_CONNECTION) {
			gameMode = GAME_MODE_MENU_LEVEL_SELECTION_MP_IN;
			cameraTargetX = GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_X;
			cameraTargetGameMode = GAME_MODE_MENU_LEVEL_SELECTION_MP;
		}
		else {
			// failed to connect / disconnecting
			networkServerDisconnect();

			gameMode = GAME_MODE_MENU_CONNECTION_OUT;
			cameraTargetX = GAME_MENU_CENTER_MENU_MULTIPLAYER_X;
			cameraTargetGameMode = GAME_MODE_MENU_MULTIPLAYER;
		}

		break;

		// -----------------------------------------------------------------------
		// LEVEL SELECTION MP
		// -----------------------------------------------------------------------

	case GAME_MODE_MENU_LEVEL_SELECTION_MP:
		if (networkServerConnectionStatus == NETWORK_SERVER_GAME) {
			// and so does the game begin...
			gameMode = GAME_MODE_MENU_GAME_MP_IN;
			cameraTargetX = GAME_MENU_CENTER_MENU_GAME_MP_X;
			cameraTargetGameMode = GAME_MODE_MENU_GAME_MP;

			// reset the game
			gameGameInit();

			break;
		}

		// manage input
		if ((keyBits & INPUT_KEY_BIT_UP) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_UP) == 0) {
			menuSelectionLevelSelectionMP--;
			if (menuSelectionLevelSelectionMP < 0)
				menuSelectionLevelSelectionMP = 2;
		}
		if ((keyBits & INPUT_KEY_BIT_DOWN) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_DOWN) == 0) {
			menuSelectionLevelSelectionMP++;
			if (menuSelectionLevelSelectionMP > 2)
				menuSelectionLevelSelectionMP = 0;
		}

		// move the menu selection
		fY = menuSelectionTargetsLevelSelectionMP[menuSelectionLevelSelectionMP];

		if (menuSelectionCenterLevelSelectionMPY < fY) {
			menuSelectionCenterLevelSelectionMPY += deltaTime * 0.1f;
			if (menuSelectionCenterLevelSelectionMPY > fY)
				menuSelectionCenterLevelSelectionMPY = fY;
		}
		else if (menuSelectionCenterLevelSelectionMPY > fY) {
			menuSelectionCenterLevelSelectionMPY -= deltaTime * 0.1f;
			if (menuSelectionCenterLevelSelectionMPY < fY)
				menuSelectionCenterLevelSelectionMPY = fY;
		}

		if (menuSelectionCenterLevelSelectionMPY == fY) {
			// change the level?
			if (menuSelectionLevelSelectionMP == GAME_MENU_SELECTION_LEVEL_SELECTION_MP_LEVEL) {
				if ((keyBits & INPUT_KEY_BIT_LEFT) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_LEFT) == 0) {
					// play the click SFX
					audioPlaySFX(SFX_MENU_CLICK);

					menuLevelSelectionMPLevel--;
					if (menuLevelSelectionMPLevel < 0)
						menuLevelSelectionMPLevel = 7;
				}
				else if ((keyBits & INPUT_KEY_BIT_RIGHT) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_RIGHT) == 0) {
					// play the click SFX
					audioPlaySFX(SFX_MENU_CLICK);

					menuLevelSelectionMPLevel++;
					if (menuLevelSelectionMPLevel > 7)
						menuLevelSelectionMPLevel = 0;
				}
			}

			// exit main menu?
			if (((keyBits & INPUT_KEY_BIT_START) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_START) == 0) ||
					((keyBits & INPUT_KEY_BIT_A) != 0 && (gameKeyBitsOld & INPUT_KEY_BIT_A) == 0)) {
				if (menuSelectionLevelSelectionMP == GAME_MENU_SELECTION_LEVEL_SELECTION_MP_DISCONNECT) {
					// play the click SFX
					audioPlaySFX(SFX_MENU_CLICK);

					gameMode = GAME_MODE_MENU_LEVEL_SELECTION_MP_OUT;
					cameraTargetX = GAME_MENU_CENTER_MENU_CONNECTION_X;
					cameraTargetGameMode = GAME_MODE_MENU_CONNECTION;
					connectionScreenConnect = NO;
				}
				else if (menuSelectionLevelSelectionMP == GAME_MENU_SELECTION_LEVEL_SELECTION_MP_START) {
					// play the click SFX
					audioPlaySFX(SFX_MENU_CLICK);

					// send start to the server
					networkServerSendStart();
				}
			}
		}

		// manage the bar DY
		fY = menuSelectionBarDYLevelSelectionMPList[menuSelectionLevelSelectionMP];

		if (menuSelectionBarDYLevelSelectionMP < fY) {
			menuSelectionBarDYLevelSelectionMP += deltaTime * 0.1f;
			if (menuSelectionBarDYLevelSelectionMP > fY)
				menuSelectionBarDYLevelSelectionMP = fY;
		}
		else if (menuSelectionBarDYLevelSelectionMP > fY) {
			menuSelectionBarDYLevelSelectionMP -= deltaTime * 0.1f;
			if (menuSelectionBarDYLevelSelectionMP < fY)
				menuSelectionBarDYLevelSelectionMP = fY;
		}

		break;

		// -----------------------------------------------------------------------
		// GAME MP
		// -----------------------------------------------------------------------

	case GAME_MODE_MENU_GAME_MP:
		gameGameIteration(deltaTime);
		break;

		// -----------------------------------------------------------------------
		// CAMERA MOVE
		// -----------------------------------------------------------------------

	case GAME_MODE_MENU_GAME_MP_IN:
	case GAME_MODE_MENU_GAME_MP_OUT:
	case GAME_MODE_MENU_CONNECTION_IN:
	case GAME_MODE_MENU_CONNECTION_OUT:
	case GAME_MODE_MENU_LEVEL_SELECTION_MP_IN:
	case GAME_MODE_MENU_LEVEL_SELECTION_MP_OUT:
		if (menuCameraX < cameraTargetX) {
			menuCameraX += deltaTime * 0.4f;
			if (menuCameraX >= cameraTargetX) {
				menuCameraX = cameraTargetX;
				gameMode = cameraTargetGameMode;
				gameTime = 0.0f;
			}
		}
		else {
			menuCameraX -= deltaTime * 0.4f;
			if (menuCameraX <= cameraTargetX) {
				menuCameraX = cameraTargetX;
				gameMode = cameraTargetGameMode;
				gameTime = 0.0f;
			}
		}

		break;
	}

	// remember the keys
	gameKeyBitsOld = keyBits;

	// scale the menu background mesh
	if (gameMode > GAME_MODE_SPLASH_5) {
		menuBGScale -= deltaTime * 0.005f;
		if (menuBGScale < 9.0f)
			menuBGScale = 9.0f;
	}

	// network I/O
	networkRead();
	networkServerSendUpdate(deltaTime);
}

// -----------------------------------------------------------------------
// draw the main menu
// -----------------------------------------------------------------------

void gameMenuDrawMain(void) {

	float pX, pY;

	// draw the selection bar
	videoSetColor(201, 166, 59, 128);

	pX = menuSelectionCenterMainX + GAME_MENU_CENTER_MENU_MAIN_X - menuCameraX;
	pY = menuSelectionCenterMainY + GAME_MENU_CENTER_MENU_MAIN_Y - menuCameraY;
	videoDrawRectangle2DWithPosition(pX, pY, -0.55f, 140, 8);

	// draw the menu items
	videoSetColor(255, 207, 60, 254);

	pX = menuItemCoordinatesMain[(GAME_MENU_SELECTION_MAIN_QUICK_TEST << 1) + 0] + GAME_MENU_CENTER_MENU_MAIN_X - menuCameraX;
	pY = menuItemCoordinatesMain[(GAME_MENU_SELECTION_MAIN_QUICK_TEST << 1) + 1] + GAME_MENU_CENTER_MENU_MAIN_Y - menuCameraY;
	videoDrawSprite(pX, pY - 24.0f, -0.5f, 0.3f, 0.3f, VIDEO_SPRITE_LOGO);
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "QUICK TEST", YES);

	pX = menuItemCoordinatesMain[(GAME_MENU_SELECTION_MAIN_MULTIPLAYER << 1) + 0] + GAME_MENU_CENTER_MENU_MAIN_X - menuCameraX;
	pY = menuItemCoordinatesMain[(GAME_MENU_SELECTION_MAIN_MULTIPLAYER << 1) + 1] + GAME_MENU_CENTER_MENU_MAIN_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "MULTIPLAYER", YES);

	pX = menuItemCoordinatesMain[(GAME_MENU_SELECTION_MAIN_CREDITS << 1) + 0] + GAME_MENU_CENTER_MENU_MAIN_X - menuCameraX;
	pY = menuItemCoordinatesMain[(GAME_MENU_SELECTION_MAIN_CREDITS << 1) + 1] + GAME_MENU_CENTER_MENU_MAIN_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "CREDITS", YES);

	pX = menuItemCoordinatesMain[(GAME_MENU_SELECTION_MAIN_SETTINGS << 1) + 0] + GAME_MENU_CENTER_MENU_MAIN_X - menuCameraX;
	pY = menuItemCoordinatesMain[(GAME_MENU_SELECTION_MAIN_SETTINGS << 1) + 1] + GAME_MENU_CENTER_MENU_MAIN_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "SETTINGS", YES);
}

// -----------------------------------------------------------------------
// draw the settings screen
// -----------------------------------------------------------------------

void gameMenuDrawSettings(void) {

	float pX, pY;

	// draw the selection bar
	videoSetColor(201, 166, 59, 128);

	pX = menuSelectionCenterSettingsX + GAME_MENU_CENTER_MENU_SETTINGS_X - menuCameraX;
	pY = menuSelectionCenterSettingsY + GAME_MENU_CENTER_MENU_SETTINGS_Y - menuCameraY;
	videoDrawRectangle2DWithPosition(pX, pY, -0.55f, 140, 8);

	// draw the menu items
	pX = menuItemCoordinatesSettings[(GAME_MENU_SELECTION_SETTINGS_GFX << 1) + 0] + GAME_MENU_CENTER_MENU_SETTINGS_X - menuCameraX;
	pY = menuItemCoordinatesSettings[(GAME_MENU_SELECTION_SETTINGS_GFX << 1) + 1] + GAME_MENU_CENTER_MENU_SETTINGS_Y - menuCameraY;

	videoSetColor(255, 207, 60, 254);
	if (videoQuality == VIDEO_QUALITY_HIGH)
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "GFX     ", YES);
	else
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "GFX    ", YES);

	videoSetColor(31, 137, 236, 254);
	if (videoQuality == VIDEO_QUALITY_HIGH)
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "    HIGH", YES);
	else
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "    LOW", YES);

	pX = menuItemCoordinatesSettings[(GAME_MENU_SELECTION_SETTINGS_MUSIC << 1) + 0] + GAME_MENU_CENTER_MENU_SETTINGS_X - menuCameraX;
	pY = menuItemCoordinatesSettings[(GAME_MENU_SELECTION_SETTINGS_MUSIC << 1) + 1] + GAME_MENU_CENTER_MENU_SETTINGS_Y - menuCameraY;

	videoSetColor(255, 207, 60, 254);
	if (audioBGM == YES)
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "MUSIC   ", YES);
	else
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "MUSIC    ", YES);

	videoSetColor(31, 137, 236, 254);
	if (audioBGM == YES)
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "      ON", YES);
	else
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "      OFF", YES);

	pX = menuItemCoordinatesSettings[(GAME_MENU_SELECTION_SETTINGS_SFX << 1) + 0] + GAME_MENU_CENTER_MENU_SETTINGS_X - menuCameraX;
	pY = menuItemCoordinatesSettings[(GAME_MENU_SELECTION_SETTINGS_SFX << 1) + 1] + GAME_MENU_CENTER_MENU_SETTINGS_Y - menuCameraY;

	videoSetColor(255, 207, 60, 254);
	if (audioSFX == YES)
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "SFX   ", YES);
	else
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "SFX    ", YES);

	videoSetColor(31, 137, 236, 254);
	if (audioSFX == YES)
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "    ON", YES);
	else
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "    OFF", YES);

	videoSetColor(255, 207, 60, 254);
	pX = menuItemCoordinatesSettings[(GAME_MENU_SELECTION_SETTINGS_BACK << 1) + 0] + GAME_MENU_CENTER_MENU_SETTINGS_X - menuCameraX;
	pY = menuItemCoordinatesSettings[(GAME_MENU_SELECTION_SETTINGS_BACK << 1) + 1] + GAME_MENU_CENTER_MENU_SETTINGS_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "BACK", YES);
}

// -----------------------------------------------------------------------
// draw the credits screen
// -----------------------------------------------------------------------

void gameMenuDrawCredits(void) {

	float pX, pY;

	// draw the menu items
	videoSetColor(255, 207, 60, 254);

	pX =   0.0f + GAME_MENU_CENTER_MENU_CREDITS_X - menuCameraX;
	pY = -14.0f + GAME_MENU_CENTER_MENU_CREDITS_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "MUSIC   JYRI LUUKKONEN", YES);

	pX =   0.0f + GAME_MENU_CENTER_MENU_CREDITS_X - menuCameraX;
	pY =  -7.0f + GAME_MENU_CENTER_MENU_CREDITS_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "ARTIST   JUKKA RAJANIEMI", YES);

	pX =   0.0f + GAME_MENU_CENTER_MENU_CREDITS_X - menuCameraX;
	pY =   0.0f + GAME_MENU_CENTER_MENU_CREDITS_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "2D GFX   LAURI RAUHANEN", YES);

	pX =   0.0f + GAME_MENU_CENTER_MENU_CREDITS_X - menuCameraX;
	pY =   7.0f + GAME_MENU_CENTER_MENU_CREDITS_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "3D GFX   OSKARI RAUNIO", YES);

	pX =   0.0f + GAME_MENU_CENTER_MENU_CREDITS_X - menuCameraX;
	pY =  14.0f + GAME_MENU_CENTER_MENU_CREDITS_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "THE REST   VILLE HELIN", YES);
}

// -----------------------------------------------------------------------
// draw the multiplayer screen
// -----------------------------------------------------------------------

void gameMenuDrawMultiplayer(void) {

	float pX, pY;

	// draw the selection bar
	videoSetColor(201, 166, 59, 128);

	pX = menuSelectionCenterMultiplayerX + GAME_MENU_CENTER_MENU_MULTIPLAYER_X - menuCameraX;
	pY = menuSelectionCenterMultiplayerY + GAME_MENU_CENTER_MENU_MULTIPLAYER_Y - menuCameraY;
	videoDrawRectangle2DWithPosition(pX, pY, -0.55f, 140, 8);

	// draw the menu items
	videoSetColor(255, 207, 60, 254);

	pX = -60.0f + menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_NAME << 1) + 0] + GAME_MENU_CENTER_MENU_MULTIPLAYER_X - menuCameraX;
	pY =   0.0f + menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_NAME << 1) + 1] + GAME_MENU_CENTER_MENU_MULTIPLAYER_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "NAME", NO);

	if (playerNameN > 0) {
		pX = -20.0f + menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_NAME << 1) + 0] + GAME_MENU_CENTER_MENU_MULTIPLAYER_X - menuCameraX;
		pY =   0.0f + menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_NAME << 1) + 1] + GAME_MENU_CENTER_MENU_MULTIPLAYER_Y - menuCameraY;
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, playerName, NO);
	}

	pX = -60.0f + menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_SERVER << 1) + 0] + GAME_MENU_CENTER_MENU_MULTIPLAYER_X - menuCameraX;
	pY =   0.0f + menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_SERVER << 1) + 1] + GAME_MENU_CENTER_MENU_MULTIPLAYER_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "SERVER", NO);

	if (serverNameN > 0) {
		pX = -20.0f + menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_SERVER << 1) + 0] + GAME_MENU_CENTER_MENU_MULTIPLAYER_X - menuCameraX;
		pY =   0.0f + menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_SERVER << 1) + 1] + GAME_MENU_CENTER_MENU_MULTIPLAYER_Y - menuCameraY;
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, serverName, NO);
	}

	pX = menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_CONNECT << 1) + 0] + GAME_MENU_CENTER_MENU_MULTIPLAYER_X - menuCameraX;
	pY = menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_CONNECT << 1) + 1] + GAME_MENU_CENTER_MENU_MULTIPLAYER_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "CONNECT", YES);

	pX = menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_BACK << 1) + 0] + GAME_MENU_CENTER_MENU_MULTIPLAYER_X - menuCameraX;
	pY = menuItemCoordinatesMultiplayer[(GAME_MENU_SELECTION_MULTIPLAYER_BACK << 1) + 1] + GAME_MENU_CENTER_MENU_MULTIPLAYER_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "BACK", YES);
}

// -----------------------------------------------------------------------
// draw the connection screen
// -----------------------------------------------------------------------

void gameMenuDrawConnection(void) {

	float pX, pY;

	// draw the menu items
	videoSetColor(255, 207, 60, 254);

	pX = GAME_MENU_CENTER_MENU_CONNECTION_X - menuCameraX;
	pY = GAME_MENU_CENTER_MENU_CONNECTION_Y - menuCameraY;

	if (connectionScreenConnect == YES)
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "CONNECTING...", YES);
	else
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "DISCONNECTING...", YES);
}

// -----------------------------------------------------------------------
// draw the multiplayer level selection screen
// -----------------------------------------------------------------------

void gameMenuDrawLevelSelectionMultiplayer(void) {

	float pX, pY;
	char tmp[8];
	int i;

	// draw the menu items
	tmp[1] = '.';
	tmp[2] = 0;
	for (i = 0; i < 4; i++) {
		pX = -60.0f + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_X - menuCameraX;
		pY = -30.0f + i*7.0f + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_Y - menuCameraY;
		tmp[0] = '1' + i;

		videoSetColor(gamePlayerColors[i*4+0] >> 1, gamePlayerColors[i*4+1] >> 1, gamePlayerColors[i*4+2] >> 1, 254);
		gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, tmp, NO);

		if (playerNameListN[i] > 0) {
			pX += 20.0f;
			gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, playerNameList[i], NO);
		}
	}

	// draw the selection bar
	videoSetColor(201, 166, 59, 128);

	pX = menuSelectionCenterLevelSelectionMPX + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_X - menuCameraX;
	pY = menuSelectionCenterLevelSelectionMPY + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_Y - menuCameraY;
	videoDrawRectangle2DWithPosition(pX, pY, -0.55f, 140, menuSelectionBarDYLevelSelectionMP);

	// draw the menu items
	videoSetColor(255, 207, 60, 254);

	pX = menuItemCoordinatesLevelSelectionMP[(GAME_MENU_SELECTION_LEVEL_SELECTION_MP_START << 1) + 0] + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_X - menuCameraX;
	pY = menuItemCoordinatesLevelSelectionMP[(GAME_MENU_SELECTION_LEVEL_SELECTION_MP_START << 1) + 1] + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "START", YES);

	pX = menuItemCoordinatesLevelSelectionMP[(GAME_MENU_SELECTION_LEVEL_SELECTION_MP_DISCONNECT << 1) + 0] + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_X - menuCameraX;
	pY = menuItemCoordinatesLevelSelectionMP[(GAME_MENU_SELECTION_LEVEL_SELECTION_MP_DISCONNECT << 1) + 1] + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_Y - menuCameraY;
	gameDrawString(pX, pY, -0.5f, 0.2f, 0.2f, "DISCONNECT", YES);

	// draw level change triangles
	pX = menuItemCoordinatesLevelSelectionMP[(GAME_MENU_SELECTION_LEVEL_SELECTION_MP_LEVEL << 1) + 0] + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_X - menuCameraX;
	pY = menuItemCoordinatesLevelSelectionMP[(GAME_MENU_SELECTION_LEVEL_SELECTION_MP_LEVEL << 1) + 1] + GAME_MENU_CENTER_MENU_LEVEL_SELECTION_MP_Y - menuCameraY;
	videoDrawTriangle(pX - 40.0f, pY, -0.53f, -1.0f, 0.0f, 1.0f, 1.0f, 1.0f, -1.0f, 10.0f, 10.0f);
	videoDrawTriangle(pX + 40.0f, pY, -0.53f, 1.0f, 0.0f, -1.0f, 1.0f, -1.0f, -1.0f, 10.0f, 10.0f);

	// draw the selected level
	videoSetColor(255, 255, 255, 255);
	videoDrawSprite(pX, pY, -0.53f, 0.35f, 0.35f, spritesLevel[menuLevelSelectionMPLevel]);
}

// -----------------------------------------------------------------------
// draw the virtual keyboard
// -----------------------------------------------------------------------

void gameMenuDrawVirtualKeyboard(void) {

	float pX, pY, f;
	int i, j;

	// is the virtual keyboard out from the view?
	if (menuVirtualKeyboardCenterY >= VIRTUAL_KEYBOARD_REST_Y)
		return;

	// init the keyboard?
	if (menuVirtualKeyboardInit == NO) {
		menuVirtualKeyboardInit = YES;

		for (i = 0; i < 4; i++) {
			for (j = 0; j < 10; j++) {
				f = 5.0f;
				menuVirtualKeyboardCoordinates[i][(j << 1) + 0] = GAME_MENU_CENTER_MENU_MULTIPLAYER_X + j*f - 23.5f;
				menuVirtualKeyboardCoordinates[i][(j << 1) + 1] = GAME_MENU_CENTER_MENU_MULTIPLAYER_Y + i*f - 7.6f;
			}
		}
	}

	// draw the background
	videoSetColor(255, 255, 255, 254);

	pX = GAME_MENU_CENTER_MENU_MULTIPLAYER_X - menuCameraX;
	pY = GAME_MENU_CENTER_MENU_MULTIPLAYER_Y - menuCameraY + menuVirtualKeyboardCenterY;
	videoDrawSprite(pX, pY, -0.3f, 0.2f, 0.2f, VIDEO_SPRITE_VIRTUAL_KEYBOARD);

	// draw the selection
	i = menuVirtualKeyboardY;
	j = menuVirtualKeyboardX;
	pX = menuVirtualKeyboardCoordinates[i][(j << 1) + 0] - menuCameraX;
	pY = menuVirtualKeyboardCoordinates[i][(j << 1) + 1] - menuCameraY + menuVirtualKeyboardCenterY;
	videoDrawSprite(pX, pY, -0.29f, 0.2f, 0.2f, VIDEO_SPRITE_CURSOR);
}

// -----------------------------------------------------------------------
// draw the game
// -----------------------------------------------------------------------

int clearR = 0;
int clearG = 0;
int clearB = 0;

void gameDraw(void) {

	if (gameMode == GAME_MODE_MENU_GAME_MP) {
		// clear the screen
		videoClearScreen(7, 65, 114, 255);

		// set the projection
		videoSetFrustum(-16.0f, 16.0f, -9.0f, 9.0f, 20.0f, 3000.0f);
	}
	else {
		// clear the screen
		videoClearScreen(clearR, clearG, clearB, 255);

		// set the projection
		videoSetFrustum(-16.0f, 16.0f, -9.0f, 9.0f, 0.1f, 3000.0f);
	}

	// bind the texture
	switch (gameMode) {
	case GAME_MODE_SPLASH_1:
	case GAME_MODE_SPLASH_2:
	case GAME_MODE_SPLASH_3:
	case GAME_MODE_SPLASH_4:
	case GAME_MODE_SPLASH_5:
		videoTextureAtlasBind(VIDEO_TEXTURE_ATLAS_SPLASH);
		break;

	case GAME_MODE_SPLASH_EXIT:
	case GAME_MODE_MENU_MAIN:
	case GAME_MODE_MENU_CREDITS_IN:
	case GAME_MODE_MENU_CREDITS:
	case GAME_MODE_MENU_CREDITS_OUT:
	case GAME_MODE_MENU_SETTINGS_IN:
	case GAME_MODE_MENU_SETTINGS:
	case GAME_MODE_MENU_SETTINGS_OUT:
	case GAME_MODE_MENU_MULTIPLAYER_IN:
	case GAME_MODE_MENU_MULTIPLAYER:
	case GAME_MODE_MENU_MULTIPLAYER_OUT:
	case GAME_MODE_MENU_CONNECTION_IN:
	case GAME_MODE_MENU_CONNECTION:
	case GAME_MODE_MENU_CONNECTION_OUT:
	case GAME_MODE_MENU_LEVEL_SELECTION_MP_IN:
	case GAME_MODE_MENU_LEVEL_SELECTION_MP:
	case GAME_MODE_MENU_LEVEL_SELECTION_MP_OUT:
	case GAME_MODE_MENU_GAME_MP_IN:
	case GAME_MODE_MENU_GAME_MP_OUT:
		videoTextureAtlasBind(VIDEO_TEXTURE_ATLAS_MENU);

		// the background meshes
		videoSetColor(16, 16, 16, 255);
		videoLoadIdentity();
		videoTranslate3f(0.0f, 0.0f, -11.7f);
		videoRotate3f(0.0f, 0.0f, runTime * 0.05f);
		videoScale3f(menuBGScale, menuBGScale, 1.0f);
		videoDrawMeshPlain(verticesCursor1, verticesCursor1Triangles);

		videoSetColor(32, 32, 32, 255);
		videoLoadIdentity();
		videoTranslate3f(0.0f, 0.0f, -11.25f);
		videoRotate3f(0.0f, 0.0f, runTime * 0.10f);
		videoScale3f(menuBGScale, menuBGScale, 1.0f);
		videoDrawMeshPlain(verticesCursor1, verticesCursor1Triangles);

		videoSetColor(48, 48, 48, 255);
		videoLoadIdentity();
		videoTranslate3f(0.0f, 0.0f, -11.0f);
		videoRotate3f(0.0f, 0.0f, runTime * 0.15f);
		videoScale3f(menuBGScale, menuBGScale, 1.0f);
		videoDrawMeshPlain(verticesCursor1, verticesCursor1Triangles);

		break;

	case GAME_MODE_MENU_GAME_MP:
		videoTextureAtlasBind(VIDEO_TEXTURE_ATLAS_GAME);
		break;
	}

	// draw
	switch (gameMode) {
	case GAME_MODE_SPLASH_1:
	case GAME_MODE_SPLASH_2:
	case GAME_MODE_SPLASH_3:
	case GAME_MODE_SPLASH_4:
	case GAME_MODE_SPLASH_5:
		videoSetColor(255, 255, 255, 255);
		videoDrawSprite(GAME_MENU_CENTER_SPLASH_1_X - menuCameraX, 0, -0.5f, 0.33f, 0.33f, VIDEO_SPRITE_SPLASH_GBAX2007);
		videoDrawSprite(GAME_MENU_CENTER_SPLASH_2_X - menuCameraX, 0, -0.5f, 0.33f, 0.33f, VIDEO_SPRITE_SPLASH_TTURSAS);
		break;

	case GAME_MODE_SPLASH_EXIT:
	case GAME_MODE_MENU_MAIN:
		gameMenuDrawMain();
		break;

	case GAME_MODE_MENU_CREDITS_IN:
	case GAME_MODE_MENU_CREDITS_OUT:
		gameMenuDrawMain();
		gameMenuDrawCredits();
		break;

	case GAME_MODE_MENU_CREDITS:
		gameMenuDrawCredits();
		break;

	case GAME_MODE_MENU_SETTINGS_IN:
	case GAME_MODE_MENU_SETTINGS_OUT:
		gameMenuDrawMain();
		gameMenuDrawSettings();
		break;

	case GAME_MODE_MENU_SETTINGS:
		gameMenuDrawSettings();
		break;

	case GAME_MODE_MENU_MULTIPLAYER_IN:
	case GAME_MODE_MENU_MULTIPLAYER_OUT:
		gameMenuDrawMain();
		gameMenuDrawMultiplayer();
		break;

	case GAME_MODE_MENU_MULTIPLAYER:
		gameMenuDrawMultiplayer();
		gameMenuDrawVirtualKeyboard();
		break;

	case GAME_MODE_MENU_CONNECTION_IN:
	case GAME_MODE_MENU_CONNECTION_OUT:
		gameMenuDrawMultiplayer();
		gameMenuDrawConnection();
		gameMenuDrawLevelSelectionMultiplayer();
		break;

	case GAME_MODE_MENU_CONNECTION:
		gameMenuDrawConnection();
		break;

	case GAME_MODE_MENU_LEVEL_SELECTION_MP_IN:
	case GAME_MODE_MENU_LEVEL_SELECTION_MP_OUT:
		gameMenuDrawConnection();
		gameMenuDrawLevelSelectionMultiplayer();
		break;

	case GAME_MODE_MENU_LEVEL_SELECTION_MP:
	case GAME_MODE_MENU_GAME_MP_IN:
	case GAME_MODE_MENU_GAME_MP_OUT:
		gameMenuDrawLevelSelectionMultiplayer();
		break;

	case GAME_MODE_MENU_GAME_MP:
		gameDrawGame();
		break;
	}
}

// ---------------------------------------------------------------------------
// convert a character to a spriteID
// ---------------------------------------------------------------------------

int gameCharToSpriteID(int c) {

	c &= 0xFF;

	if (c >= '0' && c <= '9')
		return VIDEO_SPRITE_MENU_0 + c - '0';
	else if (c >= 'A' && c <= 'Z')
		return VIDEO_SPRITE_MENU_A + c - 'A';
	else if (c == ':')
		return VIDEO_SPRITE_MENU_COLON;
	else if (c == ',')
		return VIDEO_SPRITE_MENU_COMMA;
	else if (c == '.')
		return VIDEO_SPRITE_MENU_DOT;
	else if (c == '-')
		return VIDEO_SPRITE_MENU_MINUS;
	else if (c == '@')
		return VIDEO_SPRITE_MENU_ENTER;

	return -1;
}

// ---------------------------------------------------------------------------
// draw a string
// ---------------------------------------------------------------------------

void gameDrawString(float cX, float cY, float cZ, float scaleX, float scaleY, char *string, int centered) {

	int spriteID, i;
	float dX;

	if (centered == YES) {
		dX = 0.0f;
		i = 0;
		while (string[i] != 0) {
			dX += 4.5f;
			i++;
		}

		dX -= 4.5f;
		cX -= dX * 0.5f;
	}

	i = 0;
	while (string[i] != 0) {
		spriteID = gameCharToSpriteID(string[i++]);

		if (spriteID >= 0)
			videoDrawSprite(cX, cY, cZ, scaleX, scaleY, spriteID);

		cX += 4.5f;
	}
}
